from __future__ import annotations
from typing import TYPE_CHECKING, Optional
import os
from pandas import DataFrame, merge
from geopandas import GeoDataFrame, read_file
from plotly.graph_objects import Figure, Bar
if TYPE_CHECKING:
    from geoprob_pipe import GeoProbPipe


DISTINCTIVE_COLORS = [
    '(60, 180, 75)', '(230, 25, 75)', '(255, 225, 25)', '(0, 130, 200)', '(245, 130, 48)', '(145, 30, 180)',
    '(70, 240, 240)', '(240, 50, 230)', '(210, 245, 60)', '(250, 190, 212)', '(0, 128, 128)',
    '(220, 190, 255)', '(170, 110, 40)', '(255, 250, 200)', '(128, 0, 0)', '(170, 255, 195)',
    '(128, 128, 0)', '(255, 215, 180)', '(128, 128, 128)', '(255, 255, 255)']


def get_plot_order(geoprob_pipe: GeoProbPipe) -> DataFrame:
    """ Assigns color codes (indices) and color rgb-values to each stochast such that they can be plotted with the
    same color and same order.

    :param geoprob_pipe:
    :return: DataFrame with columns 'variable', 'plot_order' and 'color'
    """

    # Define plot order
    df: DataFrame = geoprob_pipe.results.df_alphas_influence_factors_and_physical_values(filter_derived=True)
    df = df[["variable", "influence_factor"]]
    df = df.groupby(['variable']).mean()  # Influence factor average value define color order
    df = df.sort_values(by=["influence_factor"], ascending=False)
    df = df.reset_index(drop=False)
    df['plot_order'] = df.index  # Result is df with columns 'variable', 'influence_factor' (average), 'plot_order'

    # Add colors to plot order
    df_colors = DataFrame({"color": DISTINCTIVE_COLORS})  # Convert to dict to df for ease of use
    df_colors['plot_order'] = df_colors.index
    df = merge(df, df_colors, how="left", on="plot_order")

    return df


def get_influence_factors_for_vak(
        geoprob_pipe: GeoProbPipe, df_invloedsfactoren: DataFrame, vak_id: int
) -> Optional[DataFrame]:
    """ Returns the influence factors for the worst result of a scenario within the vak. May be empty dataframe if
    vak has no results.

    In the past weigh it for the uittredepunt, or average is among all uittredepunten, as was done in the past. The
    choice was made to keep the actual resulting influence factors because they sum up to 100%. There is also no
    physical reason to average them, and the worst case scenario is normative for the final result anyway. """
    df_result = geoprob_pipe.results.df_beta_scenarios
    df_result = df_result[df_result["vak_id"] == vak_id]

    # Check if results for vak
    if df_result.__len__() == 0:
        return None

    df_result = df_result.sort_values(by=["beta"], ascending=True)
    worst_uittredepunt_id = df_result.iloc[0]['uittredepunt_id']
    worst_scenario_id = df_result.iloc[0]['ondergrondscenario_id']
    df = df_invloedsfactoren[
        (df_invloedsfactoren["uittredepunt_id"] == worst_uittredepunt_id) &
        (df_invloedsfactoren["ondergrondscenario_id"] == worst_scenario_id)
    ]
    return df.sort_values(by=["plot_order"])


def invloedsfactoren(geoprob_pipe: GeoProbPipe, export: bool = False) -> Figure:

    # Get data
    df_invloedsfactoren = geoprob_pipe.results.df_alphas_influence_factors_and_physical_values(filter_derived=True)

    # Add plot order
    df_plot_order = get_plot_order(geoprob_pipe=geoprob_pipe)
    df_invloedsfactoren = merge(
        df_invloedsfactoren, df_plot_order[["variable", "plot_order", "color"]], how="left", on="variable")

    # Plot data
    fig = Figure()
    gdf_vakindeling: GeoDataFrame = read_file(
        geoprob_pipe.input_data.app_settings.geopackage_filepath, layer="vakindeling")
    vakken = {row['id']: row["naam"] for index, row in gdf_vakindeling.iterrows()}
    picked_colors = {}
    added_to_legend = []
    for vak_id, vak_naam in vakken.items():

        # Get invloedsfactoren data
        df_factoren = get_influence_factors_for_vak(geoprob_pipe, df_invloedsfactoren, vak_id)

        # If not results/factors
        if df_factoren is None:
            color = f"rgb{DISTINCTIVE_COLORS[0]}"
            # Add dummy zero value to visualize no results in vak
            # noinspection PyTypeChecker
            fig.add_trace(Bar(x=[vak_naam], y=[0], name="dummy", marker_color=color, showlegend=False))
            continue

        # Plot data
        stochasten = df_factoren['variable'].unique().tolist()
        for stochast in stochasten:
            if stochast not in picked_colors.keys():
                color = f"rgb{DISTINCTIVE_COLORS[picked_colors.__len__()]}"
                picked_colors[stochast] = color
            else:
                color = picked_colors[stochast]
            show_legend = False
            if stochast not in added_to_legend:
                added_to_legend.append(stochast)
                show_legend = True
            fig.add_trace(Bar(
                x=[vak_naam],
                y=[df_factoren[df_factoren['variable'] == stochast].iloc[0]['influence_factor'] * 100],
                name=stochast, marker_color=color, showlegend=show_legend, legendgroup=stochast,
            ))

    # Layout styling
    fig.update_layout(
        barmode='stack',
        bargap=0,       # Geen ruimte tussen groepen
        bargroupgap=0,  # Geen ruimte tussen individuele bars binnen een groep
        title='Invloedsfactoren<br>'
              '<sup>Invloedsfactoren zijn van het worstcase ondergrondscenario en uittredepunt.</sup>',
        xaxis_title='Vak',
        yaxis=dict(
            title="Percentage",
        ),  # Pas dit aan naar jouw gewenste bereik
    )

    # Export
    if export:
        export_dir = geoprob_pipe.visualizations.graphs.export_dir
        os.makedirs(export_dir, exist_ok=True)
        fig.write_html(os.path.join(export_dir, f"invloedsfactoren.html"), include_plotlyjs='cdn')
        # if geoprob_pipe.software_requirements.chrome_is_installed:
        #     fig.write_image(os.path.join(export_dir, f"invloedsfactoren.png"), format="png")

    # for stochast in
    return fig



from __future__ import annotations
from InquirerPy import inquirer
from geoprob_pipe import GeoProbPipe
import sys
from geoprob_pipe.utils.validation_messages import BColors
from typing import TYPE_CHECKING, List

if TYPE_CHECKING:
    from geoprob_pipe.cmd_app.cmd import ApplicationSettings


def request_vakken_to_run() -> str:
    list_vak_nummers_int: List[int] = []
    vakken_input_is_valid = False
    while vakken_input_is_valid is False:
        vakken_input: str = inquirer.text(
            message="Specificeer welke vakken je wilt doorrekenen. Doe dit door comma-separated de vak nummers (id) "
                    "op te geven. Bijvoorbeeld '4,5,6,7'.",
        ).execute()

        # Convert to integers
        vakken_input = vakken_input.replace(" ", "")
        list_vak_nummers_str = vakken_input.split(sep=",")
        try:
            list_vak_nummers_int = [int(vak_id) for vak_id in list_vak_nummers_str]
        except ValueError:
            print(f"{BColors.WARNING}Het bestand moet of een geopackage, shapefile of geodatabase zijn. Jouw invoer "
                  f"eindigt op de extensie .{vakken_input.split(sep='.')[-1]}.{BColors.ENDC}")
            continue

        # Assure more than one vak id is given
        if list_vak_nummers_int.__len__() == 0:
            print(f"{BColors.WARNING}Het bestand moet of een geopackage, shapefile of geodatabase zijn. Jouw invoer "
                  f"eindigt op de extensie .{vakken_input.split(sep='.')[-1]}.{BColors.ENDC}")
            continue

        vakken_input_is_valid = True

    return f"vakken:{','.join([str(item) for item in list_vak_nummers_int])}"


def run_calculations(app_settings: ApplicationSettings) -> bool:
    choices_list = [
        "Ja, alles",
        "Ja, specifieke vakken",
        "Nee (applicatie sluit af)"
    ]
    choice = inquirer.select(
        message="Wil je de berekeningen uitvoeren? \n"
                "Let op: Indien je al rekenresultaten hebt, zullen deze verwijderd worden, alvorens de nieuwe "
                "resultaten worden opgeslagen. ", choices=choices_list, default=choices_list[0],
    ).execute()

    if choice == choices_list[0]:
        geoprob_pipe = GeoProbPipe(app_settings)
        geoprob_pipe.export_archive()
        print(BColors.OKBLUE, f"✅  Berekeningen zijn uitgevoerd.", BColors.ENDC)
    elif choice == choices_list[1]:
        vakken_str: str = request_vakken_to_run()
        app_settings.to_run = vakken_str
        geoprob_pipe = GeoProbPipe(app_settings)
        geoprob_pipe.export_archive()
        print(BColors.OKBLUE, f"✅  Berekeningen zijn uitgevoerd, voor vakken "
                              f"{app_settings.to_run.replace("vakken:", "")}.", BColors.ENDC)
    elif choice == choices_list[2]:
        sys.exit(f"Applicatie is afgesloten.")
    else:
        raise ValueError
    return True



from geoprob_pipe.calculations.systems.moria.limit_state_functions import (
    system_variable_setup, calc_Z_h, calc_Z_p, calc_Z_u)
from geoprob_pipe.calculations.systems.base_objects.system_calculation import (
    SystemCalculation)
from typing import List, Dict, Union, Tuple


class MORIACalculation(SystemCalculation):
    """ Vooraf gedefinieerde System Reliability Calculation voor piping met het WBI-stijghoogtemodel. """

    def __init__(
            self,
            system_variable_distributions: List[Dict],
            system_variable_correlations: List[Tuple[str, str, float]] = None,
            project_settings: Dict[str, Union[str, float, int]] = None
    ):

        super().__init__(
            distributions=system_variable_distributions, project_settings=project_settings,
            correlations=system_variable_correlations

        )
        self.given_variables_setup_function = system_variable_setup
        self.given_limit_states = [calc_Z_u, calc_Z_h, calc_Z_p]
        # self.given_combin_limit_state = calc_Z_combin



from probabilistic_library import (
    ReliabilityProject, DesignPoint, CombineProject, ReliabilityMethod,
    CombinerMethod, CombineType, Stochast, Settings)
from typing import Optional, Callable, List, Dict, Union, Tuple, cast
from geoprob_pipe.utils.validation_messages import ValidationMessages
# noinspection PyPep8Naming
from geoprob_pipe.utils.loggers import TmpAppConsoleHandler as logger


class SystemCalculation:
    """ Pre-defined system reliability calculation for parallel systems. """

    def __init__(
            self,
            distributions: List[Dict],
            correlations: List[Tuple[str, str, float]],
            project_settings: Dict[str, Union[str, float, int]],
            # For assigning in children
            limit_states: Optional[List[Callable]] = None,
            # combin_limit_state: Optional[Callable] = None,
            variables_setup_function: Optional[Callable] = None
    ):
        """

        :param distributions:
        :param correlations:
        :param project_settings: ReliabilityProject settings for the limit
            state design points.
        :param limit_states:
        :param variables_setup_function: Dummy functie waarmee variabele namen
            worden geïnitieerd.
        """
        
        # Mutable arguments
        if project_settings is None:
            project_settings = {}
        if correlations is None:
            correlations = []

        self.validation_messages = ValidationMessages()
        self.metadata = {}

        # Input arguments
        self.given_project_settings: Dict[str, Union[str, float, int]] = (
            project_settings)
        self.given_variables_setup_function: Callable = cast(
            Callable, variables_setup_function)
        self.given_limit_states: List[Callable] = cast(
            List[Callable], limit_states)
        # self.given_combin_limit_state: Callable = cast(
        #     Callable, combin_limit_state)
        self.given_distributions: List[Dict] = distributions
        self.given_correlations: List[Tuple[str, str, float]] = correlations
        # TODO Nu Should Klein: I.p.v. dict maak gebruik van
        # Distributie-objecten. Minder fout gevoelig.

        # Placeholders
        # self.project: Optional[ReliabilityProject] = None
        self.model_design_points: List[DesignPoint] = []
        # self.combine_project: Optional[CombineProject] = None
        # self.system_design_point: Optional[DesignPoint] = None

    def run(self):
        """ Performs all logic of the system reliability calculation.
        """
        self._setup_project()
        self._apply_settings()
        self._assign_variables()
        self._assign_project_correlations()
        self._generate_model_design_points()
        self._generate_system_design_point()

    def _setup_project(self):
        """ Sets up the ReliabilityProject-object. This will be used
        for all model design points.
        """
        self.project = ReliabilityProject()
        self.project.settings.reliability_method = ReliabilityMethod.form

        # Some base settings, may be overwritten through self._apply_settings
        self.project.settings.reliability_method = 'form'
        self.project.settings.variation_coefficient = 0.02
        self.project.settings.maximum_iterations = 1000
        self.project.settings.relaxation_factor = 0.75

    def _apply_settings(self):
        """
        Set up the settings of the ReliabilityProject

        Note: all supported settings can be found in probabilistic_library
        .reliability.Settings.__dir__
        """
        for attr_name, value in self.given_project_settings.items():

            if attr_name not in Settings().__dir__():
                raise ValueError(
                    f"Attribute '{attr_name}' not found in the ReliabilityCalculation.Settings-class. "
                    f"Available attributes are:\n"
                    f"{Settings().__dir__()}")

            setattr(self.project.settings, str(attr_name), value)

    def _assign_variables(self):
        self.project.model = self.given_variables_setup_function

        # Validate all system variables have a distribution provided
        system_variable_keys = _system_variable_keys(self)
        for var_item in self.project.variables:
            var_item: Stochast
            if var_item.name not in system_variable_keys:
                raise KeyError(
                    """
                               """
                    f"The system variable '{var_item.name}' has no"
                    " distribution provided in system_variable_distributions-list."
                    " Please do so before running the system.")

        for item in self.given_distributions:
            name = item['name']

            # Check if variable exists
            if self.project.variables[name] is None:
                self.validation_messages.add_warning(
                    msg=f"The variable '{name}' is unknown in the ReliabilityProject, i.e. in the given "
                        f"'system_variables_setup'-function. For now this application skips unnecessary variables. If "
                        f"the variable is necessary, revisit your 'system_variables_setup'-function and the limit state "
                        f"functions.")
                # TODO Nu Should Klein: Feedback aan gebruiker dat er validation messages zijn.
                continue

            self.project.variables[name].distribution = item['distribution_type']

            # Key-worded arguments for uniform
            if 'minimum' in item.keys():
                self.project.variables[name].minimum = item['minimum']
            if 'maximum' in item.keys():
                self.project.variables[name].maximum = item['maximum']

            # Key-worded arguments for deterministic, normal and/or log_normal
            if 'mean' in item.keys():
                self.project.variables[name].mean = item['mean']
            if 'deviation' in item.keys():
                self.project.variables[name].deviation = item['deviation']
            if 'variation' in item.keys():
                self.project.variables[name].variation = item['variation']

            # Key-worded arguments for cdf-curve
            if 'fragility_values' in item.keys():
                self.project.variables[name].fragility_values.extend(item['fragility_values'])

    def _assign_project_correlations(self):
        for correlation in self.given_correlations:
            self.project.correlation_matrix[
                correlation[0],  # Parameter name A
                correlation[1]   # Parameter name B
            ] = correlation[2]   # Correlation value between 0.0 and 1.0

    def _generate_model_design_points(self):
        for model_callable in self.given_limit_states:
            self.project.model = model_callable
            self._assign_project_correlations()
            # noinspection PyBroadException
            try:
                self.project.run()
            except Exception:
                logger.error("Run failed in _generate_model_design_points")
            design_point = self.project.design_point
            if design_point is None:
                logger.debug(f"For limit state: {model_callable}")
                raise TypeError("Design_point is NoneType, run has failed.")
            design_point.identifier = model_callable.__name__
            self.model_design_points.append(design_point)

    def _generate_system_design_point(self):
        self.combine_project = CombineProject()
        for design_point in self.model_design_points:
            self.combine_project.design_points.append(design_point)
        self.combine_project.settings.combiner_method = CombinerMethod.importance_sampling
        self.combine_project.settings.combine_type = CombineType.parallel
        self.combine_project.run()
        self.combine_project.design_point.identifier = "system"
        self.system_design_point = self.combine_project.design_point


def _system_variable_keys(self: SystemCalculation) -> List[str]:
    return_array = []
    for item in self.given_distributions:
        print(f"{item=}")
        return_array.append(item['name'])
    return return_array
    # return [item['name'] for item in self.given_distributions]
